"""Crash reporter for MPF.

Based on https://github.com/lobocv/crashreporter/tree/master (also MIT licensed).
"""
import inspect
import logging
import re
import traceback
from datetime import datetime
from pprint import pprint
from types import FunctionType, MethodType, ModuleType, BuiltinMethodType, BuiltinFunctionType
try:
    import requests
except ImportError:
    requests = None

from mpf.exceptions.base_error import BaseError
from mpf._version import __version__

REPORTING_URL = "https://crashes.missionpinball.org/submit/"


def obr_repr(obj):
    """Return obj representation if possible."""
    try:
        return repr(obj)
    # pylint: disable-msg=broad-except
    except Exception as e:
        logging.error(e)
        return 'String Representation not found'


def string_variable_lookup(tb, s):
    """Look up the value of an object in a traceback by a dot-lookup string.

    ie. "self.crash_reporter.application_name"
    Returns ValueError if value was not found in the scope of the traceback.
    :param tb: traceback
    :param s: lookup string
    :return: value of the
    """
    refs = []
    dot_refs = s.split('.')
    dot_lookup = 0
    dict_lookup = 1
    for _, ref in enumerate(dot_refs):
        dict_refs = re.findall(r'(?<=[)(?:[\'"])([^\'\"]*)(?:[\'"])(?=])', ref)
        if dict_refs:
            bracket = ref.index('[')
            refs.append((dot_lookup, ref[:bracket]))
            refs.extend([(dict_lookup, t) for t in dict_refs])
        else:
            refs.append((dot_lookup, ref))

    scope = tb.tb_frame.f_locals.get(refs[0][1], ValueError)
    if scope is ValueError:
        return scope
    for lookup, ref in refs[1:]:
        try:
            if lookup == dot_lookup:
                scope = getattr(scope, ref, ValueError)
            else:
                scope = scope.get(ref, ValueError)
        # pylint: disable-msg=broad-except
        except Exception as e:
            logging.error(e)
            scope = ValueError

        if scope is ValueError:
            return scope
        if isinstance(scope, (FunctionType, MethodType, ModuleType, BuiltinMethodType, BuiltinFunctionType)):
            return ValueError
    return scope


def get_object_references(tb, source, max_string_length=1000):
    """Find the values of referenced attributes of objects within the traceback scope.

    :param tb: traceback
    :return: list of tuples containing (variable name, value)
    """
    referenced_attr = set()
    for line in source.split('\n'):
        referenced_attr.update(set(re.findall(
            r'[A-z]+[0-9]*\.(?:[A-z]+[0-9]*\.?)+(?!\')(?:\[(?:[\'"]).*(?:[\'"])])*(?:\.[A-z]+[0-9]*)*', line)))
    referenced_attr = sorted(referenced_attr)
    info = []
    for attr in referenced_attr:
        v = string_variable_lookup(tb, attr)
        if v is not ValueError:
            ref_string = format_reference(v, max_string_length=max_string_length)
            info.append((attr, ref_string))
    return info


def get_local_references(tb, max_string_length=1000):
    """Find the values of the local variables within the traceback scope.

    :param tb: traceback
    :return: list of tuples containing (variable name, value)
    """
    if 'self' in tb.tb_frame.f_locals:
        _locals = [('self', obr_repr(tb.tb_frame.f_locals['self']))]
    else:
        _locals = []
    for k, v in tb.tb_frame.f_locals.items():
        if k == 'self':
            continue
        try:
            vstr = format_reference(v, max_string_length=max_string_length)
            _locals.append((k, vstr))
        except TypeError:
            pass
    return _locals


def format_reference(ref, max_string_length=1000):
    """Convert an object / value into a string representation to pass along in the payload.

    :param ref: object or value
    :param max_string_length: maximum number of characters to represent the object
    :return:
    """
    def _pass(*args):
        del args
    additional_info = []
    if isinstance(ref, (list, tuple, set, dict)):
        # Check for length of reference
        length = getattr(ref, '__len__', _pass)()
        if length is not None:
            additional_info.append(('length', length))

    if additional_info:
        v_str = ', '.join(['%s: %s' % a for a in additional_info] + [obr_repr(ref)])
    else:
        v_str = obr_repr(ref)

    if len(v_str) > max_string_length:
        v_str = v_str[:max_string_length] + ' ...'

    return v_str


def analyze_traceback(tb, inspection_level=None, limit=None):
    """Extract trace back information into a list of dictionaries.

    :param tb: traceback
    :return: list of dicts containing filepath, line, module, code, traceback level and source code for tracebacks
    """
    info = []
    tb_level = tb
    extracted_tb = traceback.extract_tb(tb, limit=limit)
    for ii, (filepath, line, module, code) in enumerate(extracted_tb):
        try:
            func_source, func_lineno = inspect.getsourcelines(tb_level.tb_frame)
        except OSError:
            func_source = []
            func_lineno = 0

        d = {"file": filepath,
             "error_line_number": line,
             "module": module,
             "error_code": code,
             "module_line_number": func_lineno,
             "custom_inspection": {},
             "source_code": ''}
        if inspection_level is None or len(extracted_tb) - ii <= inspection_level:
            # Perform advanced inspection on the last `inspection_level` tracebacks.
            d['source_code'] = ''.join(func_source)
            d['local_variables'] = get_local_references(tb_level)
            d['object_variables'] = get_object_references(tb_level, d['source_code'])
        tb_level = getattr(tb_level, 'tb_next', None)
        info.append(d)

    return info


def _send_crash_report(report, reporting_url):
    r = requests.post(reporting_url, json=report, timeout=60)
    if r.status_code != 200:
        print("Failed to send report. Got response code {}. Error: {}", r.status_code, r.content)
    else:
        print(r.content.decode())


def report_crash(e: BaseException, location, config):
    """Report crash."""
    if not requests:
        print("Please install the crash_reporter feature to use the MPF crash reporter.")
        return

    log = logging.getLogger("crash_reporter")
    try:
        trace = analyze_traceback(e.__traceback__)
    # pylint: disable-msg=broad-except
    except BaseException:
        log.exception(
            "Failed to generate crash report. Please report this in our forum!")
        return

    report = {
        'timestamp': str(datetime.now()),
        'location': location,
        'exception_type': str(e.__class__),
        'trace': trace,
        'version': __version__
    }

    if isinstance(e, BaseError):
        report["error_no"] = e.get_error_no()
        report["error_context"] = e.get_context()
        report["error_logger_name"] = e.get_logger_name()
    else:
        report["error_no"] = None
        report["error_context"] = None
        report["error_logger_name"] = None

    reporting_mode = config.get("mpf", {}).get("report_crashes", "ask")

    if reporting_mode == "always":
        log.info("Will report crash as \"report_crashes\" mode is \"%s\"", reporting_mode)
        _send_crash_report(report, REPORTING_URL)
    elif reporting_mode == "ask":
        while True:
            print("\n----------------------------------------------------")
            print("MPF CRASH REPORTER")
            print("----------------------------------------------------")
            print("You can choose to anonymously report the above incident to the MPF team to "
                  "let us know for this particular issue.")
            print("Enter 'show' to see exactly what will be send.")
            print("This information will be used to improve MPF (improve ergonomics and fix issues) and "
                  "may become public. You agree that we may store this data indefinitely.")
            print("You can set this permanently in your config using mpf:report_crashes: \"never\" or \"always\"")
            print("----------------------------------------------------\n")
            try:
                response = input("Do you want to report the crash to the MPF team (yes/no/show)? ")
            except KeyboardInterrupt:
                response = "no"
            if response in ("show", "s"):
                print("Full crash report to be send:\n\n")
                pprint(report, width=160)
            elif response in ("n", "no"):
                print("Will not send report.")
                return
            elif response in ("y", "yes"):
                print("Will send report.")
                _send_crash_report(report, REPORTING_URL)
                return
            else:
                print("Invalid input")
    else:
        log.info("Will not report crash as \"report_crashes\" mode is \"%s\"", reporting_mode)
        return



"""FAST Nano Serial Interface."""
# mpf/platforms/fast/communicators/net_nano.py

from packaging import version

from mpf.platforms.fast.communicators.net_neuron import FastNetNeuronCommunicator


class FastNetNanoCommunicator(FastNetNeuronCommunicator):

    """Serial communicator for FAST Nano board."""

    MIN_FW = version.parse('1.05')
    IO_MIN_FW = version.parse('1.05')
    IGNORED_MESSAGES = ['WD:P', 'TN:P']
    MAX_IO_BOARDS = 9
    MAX_SWITCHES = 108
    MAX_DRIVERS = 48
    TRIGGER_CMD = 'TN'
    DRIVER_CMD = 'DN'
    SWITCH_CMD = 'SN'

    async def configure_hardware(self):
        """Configure the hardware with CH. Not used on a Nano."""

    def _process_sa(self, msg):
        # Nano has slightly different variation of this value, get it into a format the base can process
        _, _, _, raw_switch_data = msg.split(',')
        super()._process_sa(f'00,{raw_switch_data}')

    def _process_boot_message(self, msg):
        if msg == '00':  # rebooting
            self.machine.stop("FAST NET Nano rebooted")

        if msg == '02':  # reboot done
            self._process_reboot_done(msg)
            # TODO what else? Mark all configs as dirty? Log and warn if this was unexpected?
            # TODO add ignore_reboot option to config


# mpf.tests.test_Fast_Dmd

from mpf.tests.test_Fast import TestFastBase
from mpf.tests.MpfTestCase import test_config

class TestFastDmd(TestFastBase):
    """Tests the FAST Audio Interface boards."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.serial_connections_to_mock = ['dmd']

    def get_config_file(self):
        return 'dmd.yaml'

    def create_expected_commands(self):
            self.serial_connections['dmd'].expected_commands = dict()

    def test_dmd(self):
        # This was migrated from MPF 0.56 and has not been tested with real hardware.
        # Chances are it doesn't actually work, which should be fine because I don't
        # think anyone has these DMDs. But it should be easy to finish with a real DMD.

        pass


import logging
import sys
from mpf.core.config_loader import YamlMultifileConfigLoader

from mpf.core.machine import MachineController

machine_path = sys.argv[1]

config_loader = YamlMultifileConfigLoader(machine_path, ["config.yaml"], False, False)
config = config_loader.load_mpf_config()

options = {
    'force_platform': 'smart_virtual',
    'production': False,
    'mpfconfigfile': ["mpfconfig.yaml"],
    'configfile': ["config.yaml"],
    'debug': True,
    'bcp': True,
    'no_load_cache': False,
    'platform_integration_test': False,
    'create_config_cache': True,
    'text_ui': False,
    'consoleloglevel': logging.DEBUG,
}
logging.basicConfig(level=logging.DEBUG)
machine = MachineController(options, config)
machine.run()



from mpf.tests.MpfGameTestCase import MpfGameTestCase
from mpf.core.rgb_color import RGBColor


class TestBlinkenlight(MpfGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_platform(self):
        return 'smart_virtual'

    def get_machine_path(self):
        return 'tests/machine_files/blinkenlight/'

    def test_add_color_to_one_blinkenlight(self):
        self.post_event('start_mode1')
        self.assertPlaceholderEvaluates(0, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.assertPlaceholderEvaluates(0, 'device.blinkenlights.my_blinkenlight2.num_colors')
        self.post_event('add_color_to_first_blinkenlight')
        self.assertPlaceholderEvaluates(1, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.assertPlaceholderEvaluates(0, 'device.blinkenlights.my_blinkenlight2.num_colors')

    def test_add_color_to_two_blinkenlights(self):
        self.post_event('start_mode1')
        self.assertPlaceholderEvaluates(0, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.assertPlaceholderEvaluates(0, 'device.blinkenlights.my_blinkenlight2.num_colors')
        self.post_event('add_color_to_all_blinkenlights')
        self.assertPlaceholderEvaluates(1, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.assertPlaceholderEvaluates(1, 'device.blinkenlights.my_blinkenlight2.num_colors')

    def test_remove_color_from_one_blinkenlight(self):
        self.post_event('start_mode1')
        self.post_event('add_color_to_second_blinkenlight')
        self.assertPlaceholderEvaluates(0, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.assertPlaceholderEvaluates(1, 'device.blinkenlights.my_blinkenlight2.num_colors')
        self.post_event('remove_color_from_first_blinkenlight')
        self.assertPlaceholderEvaluates(0, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.assertPlaceholderEvaluates(1, 'device.blinkenlights.my_blinkenlight2.num_colors')
        self.post_event('remove_color_from_second_blinkenlight')
        self.assertPlaceholderEvaluates(0, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.assertPlaceholderEvaluates(0, 'device.blinkenlights.my_blinkenlight2.num_colors')

    def test_remove_all_colors_from_all_blinkenlights(self):
        self.post_event('start_mode1')
        self.post_event('start_mode2')
        self.post_event('add_color_to_first_blinkenlight')
        self.post_event('add_color_to_second_blinkenlight')
        self.post_event('add_color_to_third_blinkenlight')
        self.post_event('add_color_to_all_blinkenlights')
        self.post_event('mode2_add_color_to_first_blinkenlight')
        self.assertPlaceholderEvaluates(3, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.assertPlaceholderEvaluates(2, 'device.blinkenlights.my_blinkenlight2.num_colors')
        self.assertPlaceholderEvaluates(1, 'device.blinkenlights.my_blinkenlight3.num_colors')
        self.post_event('remove_all_colors_from_all_blinkenlights')
        self.assertPlaceholderEvaluates(0, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.assertPlaceholderEvaluates(0, 'device.blinkenlights.my_blinkenlight2.num_colors')
        self.assertPlaceholderEvaluates(0, 'device.blinkenlights.my_blinkenlight3.num_colors')

    def test_remove_mode_colors_from_one_blinkenlight(self):
        self.post_event('start_mode1')
        self.post_event('start_mode2')
        self.post_event('add_color_to_first_blinkenlight')
        self.post_event('mode2_add_color_to_first_blinkenlight')
        self.post_event('mode2_add_color2_to_first_blinkenlight')
        self.assertPlaceholderEvaluates(3, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.post_event('mode2_remove_mode_colors_from_first_blinkenlight')
        self.assertPlaceholderEvaluates(1, 'device.blinkenlights.my_blinkenlight1.num_colors')

    def test_remove_mode_colors_when_mode_ends(self):
        self.post_event('start_mode1')
        self.post_event('start_mode2')
        self.post_event('add_color_to_first_blinkenlight')
        self.post_event('add_color_to_second_blinkenlight')
        self.post_event('mode2_add_color_to_first_blinkenlight')
        self.post_event('mode2_add_color2_to_first_blinkenlight')
        self.post_event('mode2_add_color_to_second_blinkenlight')
        self.assertPlaceholderEvaluates(3, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.assertPlaceholderEvaluates(2, 'device.blinkenlights.my_blinkenlight2.num_colors')
        self.post_event('stop_mode2')
        self.assertPlaceholderEvaluates(1, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.assertPlaceholderEvaluates(1, 'device.blinkenlights.my_blinkenlight2.num_colors')

    def test_flashing_cycle(self):
        self.post_event('start_mode1')
        self.post_event('add_color_to_all_blinkenlights')
        self.post_event('add_color_to_first_blinkenlight')
        self.post_event('add_color_to_second_blinkenlight')
        self.post_event('add_color_to_third_blinkenlight')

        self.assertPlaceholderEvaluates(2, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.assertPlaceholderEvaluates(2, 'device.blinkenlights.my_blinkenlight2.num_colors')
        self.assertPlaceholderEvaluates(1, 'device.blinkenlights.my_blinkenlight3.num_colors')

        blinkenlight1 = self.machine.blinkenlights['my_blinkenlight1']
        blinkenlight2 = self.machine.blinkenlights['my_blinkenlight2']
        blinkenlight3 = self.machine.blinkenlights['my_blinkenlight3']

        blue = RGBColor('blue')
        green = RGBColor('green')
        red = RGBColor('red')
        yellow = RGBColor('yellow')
        purple = RGBColor('purple')
        cyan = RGBColor('cyan')
        off = RGBColor('off')

        self.assertEqual(blue,   blinkenlight1.light._color)
        self.assertEqual(green,  blinkenlight2.light._color)
        self.assertEqual(purple, blinkenlight3.light._color)
        self.advance_time_and_run(1)
        self.assertEqual(red,    blinkenlight1.light._color)
        self.assertEqual(green,  blinkenlight2.light._color)
        self.assertEqual(off,    blinkenlight3.light._color)
        self.advance_time_and_run(1)
        self.assertEqual(off,    blinkenlight1.light._color)
        self.assertEqual(yellow, blinkenlight2.light._color)
        self.assertEqual(purple, blinkenlight3.light._color)
        self.advance_time_and_run(1)
        self.assertEqual(blue,   blinkenlight1.light._color)
        self.assertEqual(yellow, blinkenlight2.light._color)
        self.assertEqual(off,    blinkenlight3.light._color)
        self.advance_time_and_run(1)
        self.assertEqual(red,    blinkenlight1.light._color)
        self.assertEqual(green,  blinkenlight2.light._color)
        self.assertEqual(purple, blinkenlight3.light._color)
        self.advance_time_and_run(1)
        self.assertEqual(off,    blinkenlight1.light._color)
        self.assertEqual(green,  blinkenlight2.light._color)
        self.assertEqual(off,    blinkenlight3.light._color)
        self.advance_time_and_run(1)
        self.assertEqual(blue,   blinkenlight1.light._color)
        self.assertEqual(yellow, blinkenlight2.light._color)
        self.assertEqual(purple, blinkenlight3.light._color)
        self.post_event("remove_color_from_third_blinkenlight")
        self.advance_time_and_run(1)
        self.assertEqual(red,    blinkenlight1.light._color)
        self.assertEqual(yellow, blinkenlight2.light._color)
        self.assertEqual(off,    blinkenlight3.light._color)
        self.advance_time_and_run(1)
        self.assertEqual(off,    blinkenlight1.light._color)
        self.assertEqual(green,  blinkenlight2.light._color)
        self.assertEqual(off,    blinkenlight3.light._color)
        self.advance_time_and_run(1)
        self.assertEqual(blue,   blinkenlight1.light._color)
        self.assertEqual(green,  blinkenlight2.light._color)
        self.assertEqual(off,    blinkenlight3.light._color)

    def test_priority_order(self):
        self.post_event('start_mode1')
        self.post_event('start_mode2')

        blinkenlight1 = self.machine.blinkenlights['my_blinkenlight1']

        red = RGBColor('red')
        orange = RGBColor('orange')
        off = RGBColor('off')

        self.post_event('add_color_to_first_blinkenlight')
        self.post_event('mode2_add_color_to_first_blinkenlight')
        self.assertEqual(orange, blinkenlight1.light._color)
        self.advance_time_and_run(1)
        self.assertEqual(red,    blinkenlight1.light._color)
        self.advance_time_and_run(1)
        self.assertEqual(off,    blinkenlight1.light._color)

        self.post_event('remove_all_colors_from_all_blinkenlights')
        self.advance_time_and_run(1)

        self.post_event('mode2_add_color_to_first_blinkenlight')
        self.post_event('add_color_to_first_blinkenlight')
        self.assertEqual(orange, blinkenlight1.light._color)
        self.advance_time_and_run(1)
        self.assertEqual(red,    blinkenlight1.light._color)
        self.advance_time_and_run(1)
        self.assertEqual(off,    blinkenlight1.light._color)

    def test_replace_existing_color(self):
        self.post_event('start_mode1')

        blinkenlight1 = self.machine.blinkenlights['my_blinkenlight1']

        blue = RGBColor('blue')
        darkred = RGBColor('darkred')
        off = RGBColor('off')

        self.post_event('add_color_to_all_blinkenlights')
        self.assertPlaceholderEvaluates(1, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.assertEqual(blue, blinkenlight1.light._color)
        self.advance_time_and_run(1.5)
        self.assertEqual(off, blinkenlight1.light._color)
        self.advance_time_and_run(1.5)
        self.assertEqual(blue, blinkenlight1.light._color)
        self.advance_time_and_run(1.5)
        self.assertEqual(off, blinkenlight1.light._color)
        self.advance_time_and_run(1.5)

        self.post_event('add_color_to_first_blinkenlight_with_duplicate_key')
        self.assertPlaceholderEvaluates(1, 'device.blinkenlights.my_blinkenlight1.num_colors')
        self.assertEqual(darkred, blinkenlight1.light._color)
        self.advance_time_and_run(1.5)
        self.assertEqual(off, blinkenlight1.light._color)
        self.advance_time_and_run(1.5)
        self.assertEqual(darkred, blinkenlight1.light._color)
        self.advance_time_and_run(1.5)
        self.assertEqual(off, blinkenlight1.light._color)

    def test_show_with_tokens(self):
        self.post_event('start_mode2')

        blinkenlight = self.machine.blinkenlights['my_blinkenlight2']

        gray = RGBColor('gray')
        off = RGBColor('off')

        self.assertPlaceholderEvaluates(0, 'device.blinkenlights.my_blinkenlight2.num_colors')
        self.post_event('play_blinkenlight_token_show')
        self.assertPlaceholderEvaluates(1, 'device.blinkenlights.my_blinkenlight2.num_colors')
        self.assertEqual(gray, blinkenlight.light._color)
        self.advance_time_and_run(2)
        self.assertEqual(off, blinkenlight.light._color)
        self.advance_time_and_run(2)
        self.assertEqual(gray, blinkenlight.light._color)
        self.advance_time_and_run(2)
        self.assertEqual(off, blinkenlight.light._color)



"""Fast servo implementation."""
from mpf.platforms.interfaces.servo_platform_interface import ServoPlatformInterface


class FastServo(ServoPlatformInterface):

    """A servo in the FAST platform connected to a FAST Expansion Board."""

    # __slots__ = ["number", "exp_connection"]

    def __init__(self, breakout_board, port, config):
        """Initialize servo."""
        self.config = config
        self.exp_connection = breakout_board.communicator

        self.base_address = breakout_board.address
        self.servo_index = str(int(port) - 1)  # Servos are 0-indexed
        self.max_runtime = f"{config['max_runtime']:02X}"

        self.write_config_to_servo()

    def write_config_to_servo(self):
        """Send the servo configuration to the platform."""
        min_us = f"{self.config['min_us']:02X}"
        max_us = f"{self.config['max_us']:02X}"
        home_us = f"{self.config['home_us']:02X}"

        # EM:<INDEX>,1,<MAX_TIME_MS>,<MIN>,<MAX>,<NEUTRAL><CR>
        self.exp_connection.send_with_confirmation(
            f"EM@{self.base_address}:{self.servo_index},1,{self.max_runtime},{min_us},{max_us},{home_us}",
            'EM:P')

    def go_to_position(self, position):
        """Set a servo position."""
        if position < 0 or position > 1:
            raise AssertionError("Position has to be between 0 and 1")

        # MP:<INDEX>,<POSITION>,<TIME_MS><CR>
        self.exp_connection.send_and_forget(f'MP@{self.base_address}:{self.servo_index},'
                                            f'{int(position * 255):02X},{self.max_runtime}')

    def set_speed_limit(self, speed_limit):
        """Called during servo init."""

    def set_acceleration_limit(self, acceleration_limit):
        """Called during servo init."""

    def stop(self):
        """Called during shutdown."""
        # TODO: send command to go home and power off servo



"""Contains the Blinkenlight class."""
from operator import itemgetter

from mpf.core.rgb_color import RGBColor
from mpf.core.system_wide_device import SystemWideDevice
from mpf.core.delays import DelayManager
from mpf.core.device_monitor import DeviceMonitor

MYPY = False
if MYPY:   # pragma: no cover
    from typing import NoReturn     # pylint: disable-msg=cyclic-import,unused-import


@DeviceMonitor("num_colors")
class Blinkenlight(SystemWideDevice):

    """A light that alternates between several different colors."""

    config_section = 'blinkenlights'
    collection = 'blinkenlights'
    class_label = 'blinkenlight'

    def __init__(self, machine, name):
        """Initialize blinkenlight."""
        super().__init__(machine, name)

        self._colors = []
        self.delay = DelayManager(machine)
        self._color_duration = None
        self._cycle_duration = None
        self.num_colors = 0
        self._light_key = 'blinkenlight_{}'.format(self.name)     # cache the key as string operations as expensive

    def load_config(self, config: dict):
        """Load config."""
        super().load_config(config)

        self._color_duration = self.config['color_duration']
        self._cycle_duration = self.config['cycle_duration']
        if (self._color_duration is None and self._cycle_duration is None) or \
           (self._color_duration is not None and self._cycle_duration is not None):
            self.raise_config_error("Either color_duration or cycle_duration must be specified, but not both.", 1)

    @property
    def light(self):
        """Return the light this blinkenlight controls."""
        return self.config['light']

    @property
    def num_colors_in_cycle(self):
        """Similar to num_colors, but adds 1 for the "off" color if there is one between cycles."""
        return self.num_colors + (1 if self._off_between_cycles() else 0)

    def add_color(self, color, key, priority):
        """Add a color to the blinkenlight."""
        # check if this key already exists. If it does, replace it with the incoming color/priority
        self._colors = [x for x in self._colors if x[1] != key]
        self._colors.append((color, key, priority))
        self.info_log('Color %s with key %s added', color, key)
        self._update_light()

    def remove_all_colors(self):
        """Remove all colors from the blinkenlight."""
        self._colors.clear()
        self.info_log('All colors removed')
        self._update_light()

    def remove_color_with_key(self, key):
        """Remove a color with a given key from the blinkenlight."""
        old_len = len(self._colors)
        self._colors = [x for x in self._colors if x[1] != key]
        if len(self._colors) != old_len:
            self.info_log('Color removed with key %s', key)
            self._update_light()

    def _update_light(self):
        """Update the underlying light."""
        self.delay.clear()
        self.num_colors = len(self._colors)
        if self._colors:
            self._sort_colors()
        self._perform_step()

    def _sort_colors(self):
        """Sort the blinkenlights colors by their priority."""
        self._colors = sorted(self._colors, key=itemgetter(2), reverse=True)  # priority is item 2 of the tuple

    def _off_between_cycles(self):
        return (self.config['off_when_multiple'] and len(self._colors) > 1) or (len(self._colors) == 1)

    def _get_time_between_colors_ms(self):
        if self._cycle_duration:
            delay = self._cycle_duration / self.num_colors_in_cycle
        else:
            delay = self._color_duration
        if delay < 1:
            delay = 1
        return delay

    def _get_total_cycle_ms(self):
        if self._cycle_duration:
            return self._cycle_duration
        return self._color_duration * self.num_colors_in_cycle

    def _get_current_color(self):
        now = self.machine.clock.get_time()
        offset_ms = (now * 1000) % self._get_total_cycle_ms()
        color_i = int(offset_ms / self._get_time_between_colors_ms() + 0.5)
        if color_i >= self.num_colors:
            return RGBColor("off")
        return self._colors[color_i][0]

    def _perform_step(self):
        if self.num_colors == 0:
            self.light.remove_from_stack_by_key(self._light_key)
            return
        current_color = self._get_current_color()
        color_duration_ms = self._get_time_between_colors_ms()
        cycle_ms = self._get_total_cycle_ms()
        self.light.color(current_color, priority=self.config['priority'], key=self._light_key)
        delay_ms = color_duration_ms - ((self.machine.clock.get_time() * 1000) % cycle_ms) % color_duration_ms
        self.delay.add(delay_ms, self._perform_step, name="perform_step")



#!/usr/bin/python
if __name__ == "__main__":
    import unittest
    import sys
    import argparse

    parser = argparse.ArgumentParser(description='Run tests for a machine')

    parser.add_argument("machine_path", help="Path of the machine folder.")

    parser.add_argument("-v",
                        action="store_const", dest="verbosity", const=2,
                        default=1, help="Enables verbose logging")

    args = parser.parse_args()

    sys.path.insert(0, 'machine_files/' + args.machine_path + "/")

    all_tests = unittest.TestLoader().discover('machine_files/' + args.machine_path + "/")
    unittest.TextTestRunner(verbosity=args.verbosity).run(all_tests)



"""Base class for FAST serial interfaces."""
# mpf/platforms/fast/communicators/base.py

import asyncio

from packaging import version
from serial import EIGHTBITS, PARITY_NONE, STOPBITS_ONE, SerialException

from mpf.core.logging import LogMixin
from mpf.core.utility_functions import Util

MIN_FW = version.parse('0.00')  # override in subclass


# pylint: disable-msg=too-many-instance-attributes
class FastSerialCommunicator(LogMixin):

    """Handles the serial communication to the FAST platform."""

    IGNORED_MESSAGES = []

    __slots__ = ["platform", "remote_processor", "config", "writer", "reader", "read_task", "received_msg",
                 "machine", "fast_debug", "port_debug", "remote_firmware", "send_queue", "write_task",
                 "pause_sending_until", "pause_sending_flag", "no_response_waiting", "done_waiting",
                 "ignore_decode_errors", "message_processors", "remote_model", "port", "tasks", "watchdog_cmd"]

    def __init__(self, platform, processor, config):
        """Initialize FastSerialCommunicator."""
        super().__init__()  # Initialize logging
        self.platform = platform
        self.remote_processor = processor.upper()
        self.remote_model = str()
        self.config = config
        self.writer = None
        self.reader = None
        self.tasks = list()  # higher level tasks subclasses might need
        self.read_task = None
        self.write_task = None
        self.received_msg = b''
        self.log = None
        self.machine = platform.machine
        self.fast_debug = platform.debug
        self.port = None  # string of the port we're connected to
        self.port_debug = config['debug']

        # TODO some connections have more than one processor, should there be a processor object?
        self.remote_firmware = None

        self.send_queue = asyncio.Queue()  # Tuples of ( message, pause_until_string)

        self.pause_sending_until = ''
        self.pause_sending_flag = asyncio.Event()
        self.no_response_waiting = asyncio.Event()
        self.done_waiting = asyncio.Event()
        self.no_response_waiting.set()  # Initially, we're not waiting for any response

        self.ignore_decode_errors = True

        self.message_processors = {'XX:': self._process_xx,
                                   'ID:': self._process_id}

        if config.get('watchdog', None):
            self.watchdog_cmd = f"WD:{config['watchdog']:02X}"
        else:
            self.watchdog_cmd = None

        # TODO change these to not be hardcoded
        # TODO do something with the URL endpoint
        self.configure_logging(logger=f'FAST [{self.remote_processor}]', console_level=config['debug'],
                               file_level=config['debug'], url_base='https://fastpinball.com/mpf/error')

    def __repr__(self):
        """Return representation of FAST processor."""
        return f'[FAST {self.remote_processor}]'

    async def soft_reset(self):
        """Trigger a soft reset of the serial communicator."""
        raise NotImplementedError(f"{self.__class__.__name__} does not implement soft_reset()")

    async def connect(self):
        """Connect to the port(s) for this serial interface."""
        for port in self.config['port']:
            # If this is an auto-detect that failed to detect, the port will
            # just be 'auto' and there's no reason to try and connect to it.
            if port == 'auto':
                continue
            self.log.info("Trying to connect to %s at %sbps", port, self.config['baud'])
            success = False

            while not success:
                try:
                    connector = self.machine.clock.open_serial_connection(
                        url=port, baudrate=self.config['baud'], limit=0, xonxoff=False,
                        bytesize=EIGHTBITS, parity=PARITY_NONE, stopbits=STOPBITS_ONE)
                    self.reader, self.writer = await connector
                except SerialException:
                    if not self.machine.options["production"]:
                        break

                    # if we are in production mode, retry
                    await asyncio.sleep(.1)
                    self.log.warning("Connection to port %s failed. Will retry.", port)
                else:
                    # we got a connection
                    self.log.info("Connected to %s at %sbps", port, self.config['baud'])
                    self.port = port
                    success = True
                    break

            if success:
                break
        else:
            self.log.error("Failed to connect to any of the specified ports.")
            raise SerialException(f"{self} could not connect to a serial port. Is it open in CoolTerm? ;)")

        serial = self.writer.transport.serial
        if hasattr(serial, "set_low_latency_mode"):
            try:
                serial.set_low_latency_mode(True)
                self.log.debug("Connected via low latency mode for %s.", self.config['port'])
            except (NotImplementedError, ValueError) as e:
                self.log.debug("Connected via standard mode for %s. %s", self.config['port'], e)

        # defaults are slightly high for our use case
        self.writer.transport.set_write_buffer_limits(2048, 1024)

        # read everything which is sitting in the serial
        self.writer.transport.serial.reset_input_buffer()
        # clear buffer
        # pylint: disable-msg=protected-access
        self.reader._buffer = bytearray()

        self.ignore_decode_errors = True

        await self.clear_board_serial_buffer()

        self.ignore_decode_errors = False

        self.write_task = asyncio.create_task(self._socket_writer())
        self.write_task.add_done_callback(Util.raise_exceptions)

        self.read_task = asyncio.create_task(self._socket_reader())
        self.read_task.add_done_callback(Util.raise_exceptions)

    async def clear_board_serial_buffer(self):
        """Clear out the serial buffer."""
        self.write_to_port(b'\r\r\r\r')

    async def init(self):
        """Initialize the communicator with any board-specific logic."""
        raise NotImplementedError(f"{self.__class__.__name__} does not implement init()")

    def _process_xx(self, msg):
        """Process the XX response."""
        self.log.warning("Received XX response:%s", msg)  # what are we going to do here? TODO

    def _process_id(self, msg):
        """Process the ID response."""
        processor, self.remote_model, self.remote_firmware = msg.split()

        if self.remote_processor != processor:
            self._processor_mismatch(processor)
        else:
            self.remote_processor = processor

        self.log.info("Connected to %s with firmware v%s",
                      self.remote_model, self.remote_firmware)

        if version.parse(self.remote_firmware) < MIN_FW:
            raise AssertionError(f'Firmware version mismatch. MPF requires the {self.remote_processor} processor '
                                 f'to be firmware {MIN_FW}, but yours is {self.remote_firmware}')

        self.done_processing_msg_response()

    def _processor_mismatch(self, processor):
        self.error_log(f"PORT CONFIG ERROR: You config lists port '{self.port}' for the {self.remote_processor} "
                       f"connection, but the ID: response shows that port is the {processor} connection. "
                       f"Please update your config.")
        self.machine.stop('FAST Serial port mismatch')

    # pylint: disable-msg=inconsistent-return-statements
    async def readuntil(self, separator, min_chars: int = 0):
        """Read until separator.

        Args:
        ----
            separator: Read until this separator byte.
            min_chars: Minimum message length before separator
        """
        assert self.reader is not None
        buffer = b''
        while True:
            char = await self.reader.readexactly(1)
            buffer += char
            if char == separator and len(buffer) > min_chars:
                if self.port_debug:
                    self.log.info("<<<< %s", buffer)
                return buffer

    def start_watchdog(self):
        """Start listening for commands and schedule watchdog."""
        if self.watchdog_cmd:
            self._watchdog_task()  # send one now
            self.tasks.append(self.machine.clock.schedule_interval(
                self._watchdog_task,
                self.config['watchdog'] / 2000))

    def start_tasks(self):
        """Start periodic tasks, etc.

        Called once on MPF boot, not at game start.
        """

    def stopping(self):
        """The serial connection is about to stop.

        This is called before stop() and allows you to do things that need
        to go out before the connection is closed. A 100ms delay to allow
        for this happens after this is called.
        """

    def cancel_tasks(self):
        """Cancel all outstanding tasks.

        This method is called after stopping()
        """
        for task in self.tasks:
            task.cancel()

    def stop(self):
        """Stop and shut down this serial connection."""
        self.log.debug("Stop called on serial connection %s", self.remote_processor)
        if self.read_task:
            self.read_task.cancel()
            self.read_task = None

        if self.write_task:
            self.write_task.cancel()
            self.write_task = None

        if self.writer:
            self.writer.close()
            try:
                self.machine.clock.loop.run_until_complete(self.writer.wait_closed())
            except RuntimeError as e:
                if 'Event loop stopped before Future completed.' in str(e):
                    self.log.warning("Event loop stopped before writer could close. "
                                     "This may not be an issue if the event loop was stopped intentionally.")
                else:
                    raise e
            self.writer = None

    async def send_and_wait_for_response(self, msg, pause_sending_until, log_msg=None):
        """Sends a message and awaits until the response is received.

        Parameters
        ----------
            msg (_type_): Message to send
            pause_sending_until (_type_): Response to wait for before sending the next message
            log_msg (_type_, optional): Optional version of the message that will be used in logs.
                Typically used with binary messages so the longs can contain human readable versions.
                Defaults to None which means the actual msg will be used in the logs.
        """
        await self.no_response_waiting.wait()
        self.no_response_waiting.clear()
        self.send_with_confirmation(msg, pause_sending_until, log_msg)

    # pylint: disable-msg=too-many-arguments
    async def send_and_wait_for_response_processed(self, msg, pause_sending_until, timeout=1,
                                                   max_retries=0, log_msg=None):
        """Send a message and wait for the response to be processed.

        Unlike send_and_wait_for_response(), this method will not release the wait when the response is received.
        Instead, the wait must manually be released by calling done_processing_msg_response(). This is useful for
        messages that require multiple responses, or for messages that require real processing where you don't want
        the next messages to be sent until the processing is complete.

        Parameters
        ----------
            msg (_type_): Message to send
            pause_sending_until (_type_): Response to wait for before sending the next message
            timeout (int, optional): The time (in seconds) this communicator will wait for a response.
                If a response is not received by then (based on the pause_sending_until), the message will be resent.
                Defaults to 1.
            max_retries (int, optional): How many times the message will be resent if the response is not
                received by the timeout. -1 means unlimited retries. Defaults to 0.
            log_msg (_type_, optional): Optional version of the message that will be used in logs.
                Typically used with binary messages so the longs can contain human readable versions.
                Defaults to None which means the actual msg will be used in the logs.
        """
        self.done_waiting.clear()

        retries = 0

        while max_retries == -1 or retries <= max_retries:
            try:
                await asyncio.wait_for(self.send_and_wait_for_response(msg, pause_sending_until,
                                                                       log_msg), timeout=timeout)
                break
            except asyncio.TimeoutError:
                self.log.error("Timeout waiting for response to %s. Retrying...", msg)
                retries += 1

        await self.done_waiting.wait()

    def done_processing_msg_response(self):
        """Releases the wait for the response to be processed.

        This is used in conjunction with send_and_wait_for_response_processed().
        May be called safely if there's no wait to release.
        """
        self.done_waiting.set()

    def send_with_confirmation(self, msg, pause_sending_until, log_msg=None):
        """Sends a message without blocking (returns immediately).

        Will hold future messages in the queue until the response is received.

        Parameters
        ----------
            msg (_type_): _description_
            pause_sending_until (_type_): _description_
            log_msg (_type_, optional): _description_. Defaults to None.
        """
        if log_msg:
            self.send_queue.put_nowait((f'{msg}\r'.encode(), pause_sending_until, log_msg))
        else:
            self.send_queue.put_nowait((f'{msg}\r'.encode(), pause_sending_until, msg))

    def send_and_forget(self, msg, log_msg=None):
        """Sends a message and returns immediately without waiting for a response."""
        if log_msg:
            self.send_queue.put_nowait((f'{msg}\r'.encode(), None, log_msg))
        else:
            self.send_queue.put_nowait((f'{msg}\r'.encode(), None, msg))

    def send_bytes(self, msg, log_msg):
        """Send a raw list of bytes to the communicator."""
        # Forcing log_msg since bytes are not human readable
        self.send_queue.put_nowait((msg, None, log_msg))

    def parse_incoming_raw_bytes(self, msg):
        """Parse a bytestring from the serial communicator."""
        self.received_msg += msg

        while True:
            pos = self.received_msg.find(b'\r')

            # no more complete messages
            if pos == -1:
                break

            msg = self.received_msg[:pos]
            self.received_msg = self.received_msg[pos + 1:]

            if not msg:
                continue

            try:
                msg = msg.decode()
            except UnicodeDecodeError:

                if self.machine.is_shutting_down:
                    return

                self.log.warning("Interference / bad data received: %s", msg)
                if not self.ignore_decode_errors:
                    raise

            if self.port_debug:
                self.log.info("<<<< %s", msg)

            self._dispatch_incoming_msg(msg)

    def _dispatch_incoming_msg(self, msg):
        # Figures out what to do with incoming messages
        if msg in self.IGNORED_MESSAGES:
            return

        msg_header = msg[:3]
        if msg_header in self.message_processors:
            self.message_processors[msg_header](msg[3:])
            self.no_response_waiting.set()

        # if the msg_header matches the first chars of the self.pause_sending_until, unpause sending
        # Note that the msg_header includes the colon (e.g. "DL:", "SA:") and therefore
        # pause_sending_until must also include the colon.
        if self.pause_sending_flag.is_set() and self.pause_sending_until.startswith(msg_header):
            self._resume_sending()

    def pause_sending(self, msg_header):
        """Pause the sending of serial messages until unblocked."""
        if __debug__:
            assert len(msg_header) >= 3, \
                f"Confirmation headers should be at least three characters, received '{msg_header}'"
        self.pause_sending_until = msg_header
        self.pause_sending_flag.set()

    def _resume_sending(self):
        self.pause_sending_until = None
        self.pause_sending_flag.clear()

    def _watchdog_task(self):
        """Sends the watchdog command."""
        self.send_and_forget(self.watchdog_cmd)

    async def _socket_reader(self):
        # Read coroutine
        while True:
            resp = await self.read(128)
            if resp is None:
                return
            self.parse_incoming_raw_bytes(resp)

    async def read(self, n=-1):
        """Read up to `n` bytes from the stream and log the result if debug is true.

        See :func:`StreamReader.read` for details about read and the `n` parameter.
        """
        try:
            resp = await self.reader.read(n)
        except asyncio.CancelledError:  # pylint: disable-msg=try-except-raise
            raise
        except Exception as e:  # pylint: disable-msg=broad-except
            self.log.warning("Serial error: %s", e)
            return None

        # we either got empty response (-> socket closed) or an error
        if not resp:
            self.log.warning("Serial closed.")
            self.machine.stop("Serial %s closed.", self.config["port"])
            return None

        return resp

    async def _socket_writer(self):
        # Write coroutine
        while True:
            try:
                msg, pause_sending_until, log_msg = await self.send_queue.get()

                if pause_sending_until is not None:
                    self.pause_sending(pause_sending_until)

                # Sends a message
                self.write_to_port(msg, log_msg)

                if self.pause_sending_flag.is_set():
                    await self.pause_sending_flag.wait()

            except SerialException as e:
                self.log.error(e)
                return  # TODO better way to catch shutting down?

    def write_to_port(self, msg, log_msg=None):
        """Send a message as is, without encoding or adding a <CR> character."""
        if self.port_debug:
            if log_msg:
                self.log.info(">>>> %s", log_msg)
            else:
                self.log.info(">>>> %s", msg)

        try:
            self.writer.write(msg)
        except AttributeError:
            self.log.warning("Serial connection is not open. Cannot send message: %s", msg)
            return



"""PKONE servo implementation."""
from collections import namedtuple

from mpf.platforms.interfaces.servo_platform_interface import ServoPlatformInterface

MYPY = False
if MYPY:  # pragma: no cover
    from mpf.platforms.pkone.pkone import PKONEHardwarePlatform  # pylint: disable-msg=cyclic-import,unused-import

PKONEServoNumber = namedtuple("PKONEServoNumber", ["board_address_id", "servo_number"])


class PKONEServo(ServoPlatformInterface):

    """A servo in the PKONE platform."""

    __slots__ = ["number", "platform"]

    def __init__(self, number: PKONEServoNumber, platform: "PKONEHardwarePlatform"):
        """Initialize servo."""
        self.number = number
        self.platform = platform

    def go_to_position(self, position):
        """Set a servo position."""
        if position < 0 or position > 1:
            raise AssertionError("Position has to be between 0 and 1")

        # convert from [0,1] to [0, 250]
        position_numeric = int(position * 250)

        cmd = 'PSC{}{:02d}{:03d}'.format(
            self.number.board_address_id,
            self.number.servo_number,
            position_numeric)

        self.platform.send(cmd)

    def set_speed_limit(self, speed_limit):
        """Not implemented."""

    def set_acceleration_limit(self, acceleration_limit):
        """Not implemented."""



"""PKONE Lightshow board."""
import logging
from typing import Optional

from mpf.platforms.pkone.pkone_lights import PKONELEDChannel


# pylint: disable-msg=too-many-instance-attributes
class PKONELightshowBoard:

    """PKONE Lightshow board."""

    __slots__ = ["log", "addr", "firmware_version", "hardware_rev", "rgbw_firmware", "simple_led_count", "led_groups",
                 "max_leds_per_group", "_channel_hw_drivers"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, addr, firmware_version, hardware_rev, rgbw_firmware=False):
        """Initialize PKONE Lightshow board."""
        self.log = logging.getLogger('PKONELightshowBoard {}'.format(addr))
        self.addr = addr
        self.firmware_version = firmware_version
        self.rgbw_firmware = rgbw_firmware
        self.hardware_rev = hardware_rev
        self.simple_led_count = 40  # numbers 1 - 40
        self.led_groups = 8  # numbers 1 - 8
        self.max_leds_per_group = 64  # numbers 1 - 64 for both RGB and RGBW
        self._channel_hw_drivers = {}

        for group in range(1, self.led_groups + 1):
            self._channel_hw_drivers[group] = {}

    def get_description_string(self) -> str:
        """Return description string."""
        if self.rgbw_firmware:
            return "PKONE Lightshow Board {} - RGBW Firmware: {}, Hardware Rev: {}, Simple LEDs: {}, " \
                   "RGBW LED Groups: {} (max {} LEDs per RGBW group)".format(self.addr,
                                                                             self.firmware_version,
                                                                             self.hardware_rev,
                                                                             self.simple_led_count,
                                                                             self.led_groups,
                                                                             self.max_leds_per_group)

        return "PKONE Lightshow Board {} - RGB Firmware: {}, Hardware Rev: {}, Simple LEDs: {}, " \
               "RGB LED Groups: {} (max {} LEDs per RGB group)".format(self.addr,
                                                                       self.firmware_version,
                                                                       self.hardware_rev,
                                                                       self.simple_led_count,
                                                                       self.led_groups,
                                                                       self.max_leds_per_group)

    def add_channel_hw_driver(self, group: int, channel: PKONELEDChannel):
        """Add a channel hardware driver."""
        self._channel_hw_drivers[group][channel.number] = channel

    def get_channel_hw_driver(self, group: int, number: str) -> Optional[PKONELEDChannel]:
        """Get a channel hardware driver."""
        return self._channel_hw_drivers[group].get(number, None)

    def get_all_channel_hw_drivers(self):
        """Retrieve list of all channel hardware drivers configured for use with the Lightshow board."""
        hw_drivers = []
        for group in range(1, self.led_groups + 1):
            hw_drivers.extend(list(self._channel_hw_drivers[group].values()))
        return hw_drivers



"""Perform hardware operations."""
import asyncio
import os
import random
import statistics
import time

from mpf.core.config_loader import YamlMultifileConfigLoader

from mpf.commands import MpfCommandLineParser

from mpf.core.machine import MachineController

SUBCOMMAND = True


class Command(MpfCommandLineParser):

    """Performs hardware operations."""

    def __init__(self, args, path):
        """Parse args."""
        command_name = args.pop(1)
        super().__init__(args=args, path=path)

        machine_path, remaining_args = self.parse_args()
        self.machine_path = machine_path
        self.args = remaining_args
        config_loader = YamlMultifileConfigLoader(machine_path, ["config.yaml"], True, False)

        config = config_loader.load_mpf_config()

        self.mpf = MachineController({"bcp": False,
                                      "no_load_cache": False,
                                      "platform_integration_test": False,
                                      "mpfconfigfile": os.path.join(self.mpf_path, "mpfconfig.yaml"),
                                      "configfile": ["config"],
                                      "production": False,
                                      "create_config_cache": False,
                                      "force_platform": False,
                                      "text_ui": False
                                      }, config)

        method = getattr(self, command_name)
        method()

    def scan(self):
        """Scan hardware."""
        self.mpf.clock.loop.run_until_complete(self.mpf.initialize_core_and_hardware())
        if self.mpf.thread_stopper.is_set():
            raise AssertionError("initialization failed!")

        for name, platform in self.mpf.hardware_platforms.items():
            print("{}:".format(name))
            print(platform.get_info_string())
            print("---------")

        self.mpf.shutdown()

    def firmware_update(self):
        """Upgrade firmware of platforms."""
        self.mpf.clock.loop.run_until_complete(self.mpf.initialize_core_and_hardware())
        if self.mpf.thread_stopper.is_set():
            raise AssertionError("initialization failed!")

        for name, platform in self.mpf.hardware_platforms.items():
            print("{}:".format(name))
            print(platform.update_firmware())
            print("---------")

        self.mpf.shutdown()

    # pylint: disable-msg=too-many-locals
    def _test_repeated_pulses_with_rule(self, config, pulse_ms, pause_min, pause_max):
        latency = []
        rule_latency = []
        pulse_duration = []
        rule_pulse_duration = []
        timeouts = 0

        config["flipper"].enable()

        for _ in range(100):
            # measure coil -> input latency
            pulse_start = time.time()
            config["coil1"].pulse(pulse_ms=pulse_ms)

            try:
                self.mpf.clock.loop.run_until_complete(asyncio.wait_for(
                    self.mpf.switch_controller.wait_for_switch(config["switch1"], state=1, only_on_change=False),
                    timeout=.5))
                switch_active = time.time()
                self.mpf.clock.loop.run_until_complete(asyncio.wait_for(
                    self.mpf.switch_controller.wait_for_switch(config["switch2"], state=1, only_on_change=False),
                    timeout=.5))
                switch2_active = time.time()

                self.mpf.clock.loop.run_until_complete(asyncio.wait_for(
                    self.mpf.switch_controller.wait_for_switch(config["switch1"], state=0, only_on_change=False),
                    timeout=.5))
                switch_inactive = time.time()
                self.mpf.clock.loop.run_until_complete(asyncio.wait_for(
                    self.mpf.switch_controller.wait_for_switch(config["switch2"], state=0, only_on_change=False),
                    timeout=.5))
                switch2_inactive = time.time()
            except asyncio.TimeoutError:
                print("WARNING: Ran into timeout while waiting. Check your setup!")
                timeouts += 1
                continue

            self.mpf.clock.loop.run_until_complete(asyncio.sleep(random.uniform(pause_min * 0.001, pause_max * 0.001)))

            latency.append((switch_active - pulse_start) * 1000)
            rule_latency.append((switch2_active - switch_active) * 1000)
            pulse_duration.append((switch_inactive - switch_active) * 1000)
            rule_pulse_duration.append((switch2_inactive - switch2_active) * 1000)

        print("----------------------------------------------------------------------------------------")
        print("Pulse duration: {}ms Pause: {}ms to {}ms".format(pulse_ms, pause_min, pause_max))
        print("Latency mean: {:.2f} median: {:.2f} min: {:.2f} max: {:.2f} stdev: {:.2f} variance: {:.2f}".format(
            statistics.mean(latency), statistics.median(latency), min(latency), max(latency),
            statistics.stdev(latency), statistics.variance(latency)))
        print("Rule Latency mean: {:.2f} median: {:.2f} min: {:.2f} max: {:.2f} stdev: {:.2f} variance: {:.2f}".format(
            statistics.mean(rule_latency), statistics.median(rule_latency), min(rule_latency), max(rule_latency),
            statistics.stdev(rule_latency), statistics.variance(rule_latency)))
        print("Pulse duration measured mean: {:.2f} median: {:.2f} min: {:.2f} max: {:.2f} stdev: {:.2f} "
              "variance: {:.2f}".format(
                  statistics.mean(pulse_duration), statistics.median(pulse_duration), min(pulse_duration),
                  max(pulse_duration), statistics.stdev(pulse_duration), statistics.variance(pulse_duration)))
        print("Rule Pulse duration measured mean: {:.2f} median: {:.2f} min: {:.2f} max: {:.2f} stdev: {:.2f} "
              "variance: {:.2f}".format(
                  statistics.mean(rule_pulse_duration), statistics.median(rule_pulse_duration),
                  min(rule_pulse_duration), max(rule_pulse_duration), statistics.stdev(rule_pulse_duration),
                  statistics.variance(rule_pulse_duration)))
        if timeouts:
            print("Warning: Experienced {} timeouts during benchmark. Check your setup!".format(timeouts))
        print("----------------------------------------------------------------------------------------")
        print()

        config["flipper"].disable()

    def benchmark(self):
        """Benchmark hardware."""
        self.mpf.clock.loop.run_until_complete(self.mpf.initialize())
        if self.mpf.thread_stopper.is_set():
            raise AssertionError("initialization failed!")

        print(self.mpf.switches, self.mpf.coils)
        config = self.mpf.config_validator.validate_config("hardware_benchmark",
                                                           self.mpf.config.get("hardware_benchmark", {}))
        print("1. Please confirm that you connected driver \"{}\" to switch \"{}\" and "
              "driver \"{}\" to switch \"{}\"".format(
                  config["coil1"], config["switch1"], config["coil2"], config["switch2"]))
        print("2. Turn off high voltage!")
        print("3. Hardware test will repeatedly pulse driver \"{}\" and \"{}\". "
              "Make sure they are not connected to coils and cannot cause any harm.".format(
                  config["coil1"], config["coil2"]))
        print("4. Turn off high voltage! (seriously)")
        print("")

        input_text = input("I am certain and know what I am doing (type YES if you are certain): ")
        if input_text != "YES":
            print("Wrong input. Exiting!")
            self.mpf.shutdown()
            return

        input_text = input("I did turn off high voltage (type HIGH VOLTAGE IS OFF): ")
        if input_text != "HIGH VOLTAGE IS OFF":
            print("Wrong input. Exiting!")
            self.mpf.shutdown()
            return

        print()
        print("This will take a few seconds. Please standby!")
        print()

        if config["flipper"].config["main_coil"] != config["coil2"]:
            print("Main_coil on flipper {} should be {} but is {}.".format(
                config["flipper"], config["coil2"], config["flipper"].config["main_coil"]))
            self.mpf.shutdown()
            return

        if config["flipper"].config["activation_switch"] != config["switch1"]:
            print("Activation_switch on flipper {} should be {} but is {}.".format(
                config["flipper"], config["switch1"], config["flipper"].config["activation_switch"]))
            self.mpf.shutdown()
            return

        if config["switch1"].state != 0:
            print("Switch {} should be inactive but is active.".format(config["switch1"]))
            self.mpf.shutdown()
            return

        if config["switch2"].state != 0:
            print("Switch {} should be inactive but is active.".format(config["switch2"]))
            self.mpf.shutdown()
            return

        # let the platform settle
        self.mpf.clock.loop.run_until_complete(asyncio.sleep(.5))

        self._test_repeated_pulses_with_rule(config, 53, 50, 100)
        self.mpf.clock.loop.run_until_complete(asyncio.sleep(.5))
        self._test_repeated_pulses_with_rule(config, 23, 5, 20)
        self.mpf.clock.loop.run_until_complete(asyncio.sleep(.5))
        self._test_repeated_pulses_with_rule(config, 13, 1, 2)

        self.mpf.shutdown()



"""Blinkenlight config player."""
from mpf.config_players.device_config_player import DeviceConfigPlayer
from mpf.devices.blinkenlight import Blinkenlight


class BlinkenlightPlayer(DeviceConfigPlayer):

    """Blinks lights different colors based on config."""

    config_file_section = 'blinkenlight_player'
    show_section = 'blinkenlights'
    machine_collection_name = 'blinkenlights'
    allow_placeholders_in_keys = True

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Adds/removes colors to blink."""
        del kwargs

        for blinkenlight, s in settings.items():
            if not isinstance(blinkenlight, str):
                action = s['action']
                key = context + s['key'] if s["key"] else context
                if action == 'add':
                    self._add_color(blinkenlight, s['color'], key, priority + s['priority'])
                    self._get_instance_dict(context)[(key, blinkenlight)] = blinkenlight
                elif action == 'remove':
                    self._remove_color(blinkenlight, key)
                    try:
                        del self._get_instance_dict(context)[(key, blinkenlight)]
                    except KeyError:
                        pass
                elif action == 'remove_mode':
                    self.clear_context(context)
                elif action == 'remove_all':
                    self._remove_all_colors(blinkenlight)
                else:
                    raise AssertionError("Unknown action {}".format(action))

    @staticmethod
    def _add_color(blinkenlight: Blinkenlight, color, key, priority):
        """Instructs a blinkenlight to add a color to its list of colors."""
        blinkenlight.add_color(color, key, priority)

    @staticmethod
    def _remove_all_colors(blinkenlight):
        """Instructs a blinkenlight to remove all of its colors."""
        blinkenlight.remove_all_colors()

    @staticmethod
    def _remove_color(blinkenlight, key):
        """Instructs a blinkenlight to remove a color with a given key from its list of colors."""
        blinkenlight.remove_color_with_key(key)

    def clear_context(self, context):
        """Clear the context. In our case, this means remove the mode colors from all blinkenlights."""
        for (key, _), blinkenlight in self._get_instance_dict(context).items():
            blinkenlight.remove_color_with_key(key)
        self._reset_instance_dict(context)

    def get_express_config(self, value):
        """Parse express config."""
        if value in ("stop", "remove"):
            return dict(action="remove")
        return dict(color=value)



"""A coil/driver in the PKONE platform."""
import logging
from collections import namedtuple
from typing import Optional

from mpf.core.platform import DriverConfig, SwitchSettings
from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface, PulseSettings, HoldSettings

MYPY = False
if MYPY:  # pragma: no cover
    from mpf.platforms.pkone.pkone import PKONEHardwarePlatform  # pylint: disable-msg=cyclic-import,unused-import

PKONECoilNumber = namedtuple("PKONECoilNumber", ["board_address_id", "coil_number"])
PKONECoilConfiguration = namedtuple("PKONECoilConfiguration", ["pulse_settings", "hold_settings", "recycle_time"])


class PKONECoil(DriverPlatformInterface):

    """Base class for coils/drivers connected to a PKONE Controller/Extension."""

    __slots__ = ["log", "hardware_rule", "_config_state", "machine", "platform",
                 "send", "platform_settings"]

    def __init__(self, config: DriverConfig, platform: "PKONEHardwarePlatform", number: PKONECoilNumber,
                 platform_settings: dict) -> None:
        """Initialize Coil."""
        super().__init__(config, number)
        self.log = logging.getLogger('PKONECoil')
        self.hardware_rule = False  # type: bool
        self._config_state = None  # type: Optional[PKONECoilConfiguration]
        self.machine = platform.machine
        self.platform = platform
        self.send = platform.controller_connection.send
        self.platform_settings = platform_settings

        self.log.debug("Initialize Coil Settings: %s", self.config)
        self.reset()

    def get_board_name(self):
        """Return PKONE Extension addr."""
        if self.number.board_address_id not in self.platform.pkone_extensions.keys():
            return "PKONE Unknown Board"
        return "PKONE Extension Board {}".format(self.number.board_address_id)

    def get_recycle_time_ms_for_cmd(self, recycle, pulse_ms) -> int:
        """Return recycle ms."""
        if not recycle:
            return 0
        if self.platform_settings['recycle_ms'] is not None:
            return self.platform_settings['recycle_ms']

        # default recycle_ms to pulse_ms * 2 (cap at 500ms)
        return min(pulse_ms * 2, 500)

    def reset(self):
        """Reset a coil - removes all settings, including any hardware rules, for the coil."""
        self.log.debug("Resetting coil %s", self.number)

        # Ensure settings will be sent to the coil before it is activated again
        self._config_state = None

        if self.hardware_rule:
            self.clear_hardware_rule()

        # Configure coil with default/empty/0 settings (coil will not fire until it is configured with
        # real settings values)
        cmd = "PCC{}{:02d}0000000000".format(self.number.board_address_id, self.number.coil_number)
        self.send(cmd)

    def configure_coil(self, pulse_settings: PulseSettings, hold_settings: Optional[HoldSettings],
                       recycle_time: int = 0) -> None:
        """Configure a coil (will overwrite existing configuration settings)."""
        new_config_state = PKONECoilConfiguration(pulse_settings, hold_settings, recycle_time)

        # if config would not change do nothing
        if new_config_state == self._config_state:
            return

        # send the new coil configuration
        self._config_state = new_config_state
        cmd = "PCC{}{:02d}{:03d}{:02d}{:02d}{:03d}".format(self.number.board_address_id,
                                                           self.number.coil_number,
                                                           pulse_settings.duration,
                                                           # power must be mapped from 0-1 to 0-99
                                                           int(pulse_settings.power * 99),
                                                           int(hold_settings.power * 99) if hold_settings else 0,
                                                           recycle_time)
        self.log.debug("Sending Configure Coil command: %s", cmd)
        self.send(cmd)

    def disable(self) -> None:
        """Disable/release (turn off) this coil."""
        cmd = "PCR{}{:02d}".format(self.number.board_address_id, self.number.coil_number)
        self.log.debug("Sending Release/Disable Coil command: %s", cmd)
        self.send(cmd)

    # pylint: disable-msg=too-many-arguments
    def set_hardware_rule(self, mode: int, switch_settings: SwitchSettings,
                          eos_switch_settings: Optional[SwitchSettings], delay_time: int,
                          pulse_settings: PulseSettings, hold_settings: Optional[HoldSettings]) -> None:
        """Set a hardware rule for an autofire coil."""
        self.hardware_rule = True

        cmd = "PHR{}{:02d}{}{:02d}{}{:02d}{}" \
              "{:03d}{:03d}{:02d}{:02d}{:03d}".format(self.number.board_address_id,
                                                      self.number.coil_number,
                                                      mode,
                                                      switch_settings.hw_switch.number.switch_number,
                                                      1 if switch_settings.invert else 0,
                                                      eos_switch_settings.hw_switch.number.switch_number
                                                      if eos_switch_settings else 0,
                                                      1 if eos_switch_settings and eos_switch_settings.invert else 0,
                                                      delay_time,
                                                      pulse_settings.duration,
                                                      int(pulse_settings.power * 99),
                                                      int(hold_settings.power * 99) if hold_settings else 0,
                                                      self.get_recycle_time_ms_for_cmd(self.config.default_recycle,
                                                                                       pulse_settings.duration))

        self.log.debug("Writing Hardware Rule for coil: %s", cmd)
        self.send(cmd)

    def clear_hardware_rule(self) -> None:
        """Clear hardware rule."""
        cmd = "PHD{}{:02d}".format(self.number.board_address_id, self.number.coil_number)
        self.log.debug("Clearing Hardware Rule for coil: %s", cmd)
        self.send(cmd)
        self.hardware_rule = False

    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings) -> None:
        """Enable (turn on) this coil/driver."""
        # reconfigure coil (if necessary)
        self.configure_coil(pulse_settings, hold_settings,
                            self.get_recycle_time_ms_for_cmd(self.config.default_recycle, pulse_settings.duration))
        cmd = "PCH{}{:02d}".format(self.number.board_address_id, self.number.coil_number)
        self.log.debug("Sending Hold/Enable coil command: %s", cmd)
        self.send(cmd)

    def pulse(self, pulse_settings: PulseSettings) -> None:
        """Pulse this coil/driver."""
        # reconfigure coil (if necessary) -- use existing hold settings (if they exist) as the
        # hold settings do not matter for a pulse command (may keep the coil configuration from
        # having to be rewritten)
        hold_settings = self._config_state.hold_settings if self._config_state else None
        self.configure_coil(pulse_settings, hold_settings,
                            self.get_recycle_time_ms_for_cmd(self.config.default_recycle, pulse_settings.duration))

        # pulse/trigger coil
        cmd = "PCP{}{:02}".format(self.number.board_address_id, self.number.coil_number)
        self.log.debug("Sending Pulse coil command: %s", cmd)
        self.send(cmd)

    def timed_enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Pulse and enable the coil for an explicit duration."""
        raise NotImplementedError



"""FAST Expansion Board Serial Communicator."""
# mpf/platforms/fast/communicators/exp.py

from mpf.platforms.fast.fast_defines import EXPANSION_BOARD_FEATURES
from mpf.platforms.fast.fast_exp_board import FastExpansionBoard
from mpf.platforms.fast.communicators.base import FastSerialCommunicator

from mpf.core.utility_functions import Util

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class FastExpCommunicator(FastSerialCommunicator):

    """Handles the serial communication for the FAST EXP bus."""

    IGNORED_MESSAGES = ['XX:F']

    __slots__ = ["exp_boards_by_address", "active_board"]

    def __init__(self, platform, processor, config):
        """Initialize the EXP communicator."""
        super().__init__(platform, processor, config)

        self.exp_boards_by_address = dict()  # keys = board addresses, values = FastExpansionBoard objects
        self.active_board = None
        self.message_processors['BR:'] = self._process_br

    async def init(self):
        """Query the expansion boards."""
        await self.query_exp_boards()

    def start_tasks(self):
        """Start listening for commands and schedule watchdog."""
        for board in self.exp_boards_by_address.values():
            self.tasks.append(self.platform.machine.clock.schedule_interval(
                              board.update_leds, 1 / board.config['led_hz']))

    def stopping(self):
        """Stop listening to the board and clear it."""
        for board in self.exp_boards_by_address.values():
            board.communicator.send_and_forget(f'BR@{board.address}:')

    async def soft_reset(self):
        """Trigger a soft reset for the board and all breakouts."""
        for board in self.exp_boards_by_address.values():
            await board.soft_reset()

    async def query_exp_boards(self):
        """Query the EXP bus for connected boards."""
        for board_name, board_config in self.config['boards'].items():

            # FP-eXp-0071-2 -> FP-EXP-0071
            board_config['model'] = ('-').join(board_config['model'].split('-')[:3]).upper()

            if board_config['address']:  # need to do it this way since valid config will have 'address' = None
                board_address = board_config['address']
            else:
                board_address = EXPANSION_BOARD_FEATURES[board_config['model']]['default_address']

            # Got an ID for a board that's already registered. This shouldn't happen?
            if board_address in self.exp_boards_by_address:
                raise AssertionError(f'Expansion Board at address {board_address} is already registered')

            board_obj = FastExpansionBoard(board_name, self, board_address, board_config)
            self.exp_boards_by_address[board_address] = board_obj  # registers with this EXP communicator
            self.platform.register_expansion_board(board_obj)  # registers with the platform

            self.active_board = board_address
            await self.send_and_wait_for_response_processed(f'ID@{board_address}:', 'ID:')

            for breakout_board in board_obj.breakouts.values():
                self.active_board = breakout_board.address
                await self.send_and_wait_for_response_processed(f'ID@{breakout_board.address}:', 'ID:')

            await board_obj.reset()

    def _process_id(self, msg: str):
        self.exp_boards_by_address[self.active_board[:2]].verify_hardware(msg, self.active_board)
        self.active_board = None
        self.done_processing_msg_response()

    def _process_br(self, msg):
        del msg
        self.active_board = None
        self.done_processing_msg_response()

    def set_led_fade_rate(self, board_address: str, rate: int) -> None:
        """Sets the hardware LED fade rate for an EXP board.

        Parameters
        ----------
            board_address (str): 2 hex character board address
            rate (int): Fade rate, in milliseconds, between 0 and 8191

        Raises
        ------
            ValueError: If the fade rate is out of bounds
        """
        if not 0 <= rate <= 8191:
            raise ValueError(f"FAST LED fade rate of {rate}ms is out of bounds. Must be between 0 and 8191ms")

        self.platform.debug_log("%s - Setting LED fade rate to %sms", self, rate)
        self.send_and_forget(f'RF@{board_address}:{Util.int_to_hex_string(rate, True)}')



"""Manager for segment display text transitions."""

from mpf.devices.segment_display.transitions import (PushTransition, CoverTransition,
                                                     UncoverTransition, WipeTransition, SplitTransition)

TRANSITIONS = {
    "push": PushTransition,
    "cover": CoverTransition,
    "uncover": UncoverTransition,
    "wipe": WipeTransition,
    "split": SplitTransition
}


class TransitionManager:

    """Manages segment display text transitions."""

    __slots__ = []

    @staticmethod
    def get_transition(output_length: int, collapse_dots: bool, collapse_commas: bool, use_dots_for_commas: bool,
                       transition_config=None):
        """Create a transition instance based on the specified configuration."""
        if transition_config:
            config = transition_config.copy()
            config.pop('type')
            return TRANSITIONS[transition_config['type']](output_length, collapse_dots, collapse_commas,
                                                          use_dots_for_commas, config)

        return None

    @staticmethod
    def validate_config(config, config_validator):
        """Validate segment display transition config."""
        if 'transition' in config and config['transition']:
            if not isinstance(config['transition'], dict):
                config['transition'] = dict(type=config['transition'])

            try:
                config['transition'] = (
                    config_validator.validate_config(
                        'segment_display_transitions:{}'.format(config['transition']['type']),
                        config['transition']))

            except KeyError:
                raise ValueError('transition: section of config requires a "type:" setting')
        else:
            config['transition'] = None

        if 'transition_out' in config and config['transition_out']:
            if not isinstance(config['transition_out'], dict):
                config['transition_out'] = dict(type=config['transition_out'])

            try:
                config['transition_out'] = (
                    config_validator.validate_config(
                        'segment_display_transitions:{}'.format(
                            config['transition_out']['type']),
                        config['transition_out']))

            except KeyError:
                raise ValueError('transition_out: section of config requires a "type:" setting')
        else:
            config['transition_out'] = None

        return config



"""Base class for MPF Platform Integration Tests."""

from mpf.core.logging import LogMixin

class MpfPlatformIntegrationTest(LogMixin):

    __slots__ = ("machine", "runner")

    test_start_event = "mode_attract_started"
    initial_switches = None

    def __init__(self, runner):
        super().__init__()
        self.runner = runner
        self.machine = runner.machine
        self.configure_logging(type(self).__name__)

    def run_test(self):
        """This method is called when it's time to run the test."""
        raise NotImplementedError



"""RGB Board Serial Communicator."""
# mpf/platforms/fast/communicators/rgb.py

from packaging import version

from mpf.platforms.fast.communicators.base import FastSerialCommunicator

MIN_FW = version.parse('0.87')  # override in subclass


class FastRgbCommunicator(FastSerialCommunicator):

    """Handles the serial communication for legacy FAST RGB processors.

    Includes the Nano Controller and FP-EXP-0800 LED controller.
    """

    IGNORED_MESSAGES = ['RX:P']

    def __init__(self, platform, processor, config):
        """Initialize the RGB platform and process boot message."""
        super().__init__(platform, processor, config)

        self.message_processors['!B:'] = self._process_boot_msg

    async def init(self):
        """Initialize the RGB processor and await an ID response."""
        # Loop here until we get a response
        await self.send_and_wait_for_response_processed('ID:', 'ID:', max_retries=-1)

    def _process_boot_msg(self, msg):
        """Process bootloader message."""
        self.log.debug("Got Bootloader message: !B:%s", msg)
        if msg in ('00', '02'):
            if self.config['ignore_reboot']:
                self.machine.events.post("fast_rgb_rebooted", msg=msg)
                self.log.error("FAST RGB processor rebooted. Ignoring.")
            else:
                self.log.error("FAST RGB processor rebooted.")
                self.machine.stop("FAST RGB processor rebooted")

    def update_leds(self):
        """Update all the LEDs connected to the RGB processor of a FAST Nano controller.

        This is done once per game loop for efficiency (i.e. all LEDs are sent as a single
        update rather than lots of individual ones).

        """
        dirty_leds = [led for led in self.platform.fast_rgb_leds.values() if led.dirty]

        if dirty_leds:
            msg = 'RS:' + ','.join(["%s%s" % (led.number, led.current_color) for led in dirty_leds])
            self.send_and_forget(msg)

    def start_tasks(self):
        """Start listening for commands and schedule watchdog."""
        self.reset()

        if self.config['led_hz'] > 30:
            self.config['led_hz'] = 30

        self.tasks.append(self.machine.clock.schedule_interval(
                          self.update_leds, 1 / self.config['led_hz']))

    async def soft_reset(self, **kwargs):
        """Reset the NET processor."""
        del kwargs
        #await self.send_and_wait_for_response('RA:000000', 'RX:P')
        self.send_and_forget("RA:000000")

    def reset(self):
        """Reset the RGB processor."""
        # self.send_and_forget('RF:0')  # TODO confirm if the RGB supports RF. I think not?
        self.send_and_forget('RA:000000')
        # self.send_and_forget(f"RF:{Util.int_to_hex_string(self.config['led_fade_time'])}")

    def stopping(self):
        """Stop the RGB processor and reset it."""
        self.reset()
